package org.dripdronescanner.android.app;

import android.Manifest;
import android.content.Context;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;
import androidx.fragment.app.Fragment;

import org.dripdronescanner.android.R;
import org.dripdronescanner.android.data.AircraftObject;
import org.osmdroid.config.Configuration;
import org.osmdroid.tileprovider.tilesource.TileSourceFactory;
import org.osmdroid.util.GeoPoint;
import org.osmdroid.views.MapView;

import java.util.ArrayList;
import java.util.HashMap;

import android.Manifest;
import android.content.pm.PackageManager;
import android.graphics.Color;
import android.os.Bundle;
import android.util.Log;
import android.util.Pair;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import androidx.annotation.Nullable;
import androidx.core.app.ActivityCompat;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModelProviders;

import com.google.android.gms.internal.maps.zzt;
import com.google.android.gms.maps.CameraUpdateFactory;
import com.google.android.gms.maps.GoogleMap;
import com.google.android.gms.maps.OnMapReadyCallback;
import com.google.android.gms.maps.SupportMapFragment;
import com.google.android.gms.maps.model.BitmapDescriptorFactory;
import com.google.android.gms.maps.model.CameraPosition;
import com.google.android.gms.maps.model.LatLng;
import com.google.android.gms.maps.model.Marker;
import com.google.android.gms.maps.model.MarkerOptions;
//import com.google.android.gms.maps.model.Polyline;
import com.google.android.gms.maps.model.PolylineOptions;

import org.dripdronescanner.android.data.AircraftObject;
import org.dripdronescanner.android.data.LocationData;
import org.dripdronescanner.android.data.SystemData;
import org.dripdronescanner.android.data.Util;
import org.osmdroid.views.overlay.Polyline;
import org.osmdroid.views.overlay.mylocation.GpsMyLocationProvider;
import org.osmdroid.views.overlay.mylocation.MyLocationNewOverlay;

import java.util.Collection;
import java.util.HashMap;

public class AircraftMapView extends Fragment {
    private MapView map = null;
    private static final String TAG = "AircraftMapView";
    private AircraftViewModel model;
    private final HashMap<AircraftObject, MapObserver> aircraftObservers = new HashMap<>();
    private MyLocationNewOverlay locationOverlay;

    private final Util.DiffObserver<AircraftObject> allAircraftObserver = new Util.DiffObserver<AircraftObject>() {
        @Override
        public void onAdded(Collection<AircraftObject> added) {
            for (AircraftObject aircraftObject : added) {
                trackAircraft(aircraftObject);
            }
        }

        @Override
        public void onRemoved(Collection<AircraftObject> removed) {
            for (AircraftObject aircraftObject : removed) {
                stopTrackingAircraft(aircraftObject);
            }
        }
    };

    private void trackAircraft(AircraftObject aircraftObject) {
        MapObserver observer = new MapObserver(aircraftObject);
        aircraftObservers.put(aircraftObject, observer);
    }

    private void stopTrackingAircraft(AircraftObject aircraftObject) {
        MapObserver observer = aircraftObservers.remove(aircraftObject);
        if (observer == null) return;
        observer.stop();
    }

    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
        Context ctx = getContext();
        Configuration.getInstance().load(ctx, PreferenceManager.getDefaultSharedPreferences(ctx));

        View v = inflater.inflate(R.layout.activity_main, null);
        map = v.findViewById(R.id.map_view);
        map.setTileSource(TileSourceFactory.MAPNIK);

        map.addOnFirstLayoutListener(new MapView.OnFirstLayoutListener() {
            @Override
            public void onFirstLayout(View v, int left, int top, int right, int bottom) {
                if (getActivity() == null)
                    return;

                if (ActivityCompat.checkSelfPermission(getActivity(), Manifest.permission.ACCESS_FINE_LOCATION) != PackageManager.PERMISSION_GRANTED &&
                        ActivityCompat.checkSelfPermission(getActivity(), Manifest.permission.ACCESS_COARSE_LOCATION) != PackageManager.PERMISSION_GRANTED) {
                    // TODO: Consider calling ActivityCompat#requestPermissions
                    // to request the missing permissions, and then overriding
                    //   public void onRequestPermissionsResult(int requestCode, String[] permissions,
                    //                                          int[] grantResults)
                    // to handle the case where the user grants the permission. See the documentation
                    // for ActivityCompat#requestPermissions for more details.
                    Log.e("XX", "##################### can't make the right permissions");
                    return;
                }

                locationOverlay = new MyLocationNewOverlay(new GpsMyLocationProvider(getContext()), map);
                locationOverlay.enableMyLocation();
                locationOverlay.enableFollowLocation();
                map.getOverlays().add(locationOverlay);

                map.setMultiTouchControls(true);

                //map.getController().setCenter(new GeoPoint(58.410809, 15.621373));
                map.getController().setZoom(DESIRED_ZOOM);
            }
        });
        return v;
    }

    @Override
    public void onResume() {
        super.onResume();
        //this will refresh the osmdroid configuration on resuming.
        //if you make changes to the configuration, use
        //SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
        //Configuration.getInstance().load(this, PreferenceManager.getDefaultSharedPreferences(this));
        map.onResume(); //needed for compass, my location overlays, v6.0.0 and up
    }

    @Override
    public void onPause() {
        super.onPause();
        //this will refresh the osmdroid configuration on resuming.
        //if you make changes to the configuration, use
        //SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(this);
        //Configuration.getInstance().save(this, prefs);
        map.onPause();  //needed for compass, my location overlays, v6.0.0 and up
    }

    @Override
    public void onActivityCreated(Bundle bundle) {
        super.onActivityCreated(bundle);
        setupModel();
    }

    private static final double DESIRED_ZOOM = 17;
    private static final int ALLOWED_ZOOM_MARGIN = 2;

    private void setupModel() {
        if (getActivity() == null)
            return;

        model = ViewModelProviders.of(getActivity()).get(AircraftViewModel.class);

        model.getAllAircraft().observe(getViewLifecycleOwner(), allAircraftObserver);
        model.getActiveAircraft().observe(getViewLifecycleOwner(), new Observer<AircraftObject>() {
            MapObserver last = null;

            @Override
            public void onChanged(@Nullable AircraftObject object) {
                if (object == null || object.getLocation() == null || map == null)
                    return;
                MapObserver observer = aircraftObservers.get(object);
                if (observer == null)
                    return;

                if (object.getLocation().getLatitude() == 0.0 && object.getLocation().getLongitude() == 0.0)
                    return;

                LatLng ll = new LatLng(object.getLocation().getLatitude(), object.getLocation().getLongitude());
                Log.i(TAG, "centering on " + object + " at " + ll);

                if (last != null && last.marker != null) {
                    last.marker.setAlpha(0.5f);
                    if (last.markerPilot != null)
                        last.markerPilot.setAlpha(0.5f);
                }
                if (observer.marker != null)
                    observer.marker.setAlpha(1.0f);
                if (observer.markerPilot != null)
                    observer.markerPilot.setAlpha(1.0f);

                last = observer;

                //CameraPosition position = googleMap.getCameraPosition();
                double currentZoom = map.getZoomLevelDouble();
                if (currentZoom < DESIRED_ZOOM - ALLOWED_ZOOM_MARGIN || currentZoom > DESIRED_ZOOM + ALLOWED_ZOOM_MARGIN) {
                    // googleMap.moveCamera(CameraUpdateFactory.newLatLngZoom(ll, DESIRED_ZOOM));
                    //map.getController().animateTo(new GeoPoint(ll.latitude, ll.longitude));
                } else {
                    //map.getController().animateTo(new GeoPoint(ll.latitude, ll.longitude));
                    //googleMap.moveCamera(CameraUpdateFactory.newLatLng(ll));
                }
            }
        });
    }

    /*@Override
    public boolean onMarkerClick(Marker marker) {

        if (marker != null) {
            Object tag = marker.getTag();
            if (tag instanceof AircraftObject) {
                model.setActiveAircraft((AircraftObject) tag);
                return true;
            }
        }
        return false;
    }*/

    class MapObserver implements Observer<LocationData> {
        private Marker marker;
        private Marker markerPilot;
        private Polyline polyline;
        private PolylineOptions polylineOptions;

        private final AircraftObject aircraft;

        MapObserver(AircraftObject active) {
            aircraft = active;
            aircraft.location.observe(AircraftMapView.this, this);
            aircraft.system.observe(AircraftMapView.this, systemObserver);
            polylineOptions = new PolylineOptions()
                    .color(Color.RED)
                    .clickable(true);
        }

        void stop() {
            aircraft.location.removeObserver(this);
            aircraft.system.removeObserver(systemObserver);
            if (marker != null) {
                marker.remove();
                marker = null;
            }
            if (markerPilot != null) {
                markerPilot.remove();
                markerPilot = null;
            }
            if (polyline != null) {
                //polyline.remove();
                polyline = null;
            }
            polylineOptions = null;
        }

        private final Observer<SystemData> systemObserver = new Observer<SystemData>() {
            @Override
            public void onChanged(@Nullable SystemData ignore) {
                SystemData sys = aircraft.getSystem();
                if (sys == null || map == null)
                    return;

                // filter out zero data
                if (sys.getOperatorLatitude() == 0.0 && sys.getOperatorLongitude() == 0.0)
                    return;

                LatLng latLng = new LatLng(sys.getOperatorLatitude(), sys.getOperatorLongitude());
                if (markerPilot == null) {
                    String id = "ID missing";
                    if (aircraft.getIdentification() != null)
                        id = aircraft.getIdentification().getUasIdAsString();

                    Marker pilotMarker = new Marker((zzt) map);
                    pilotMarker.setPosition(latLng);
                    pilotMarker.setIcon(BitmapDescriptorFactory.defaultMarker(BitmapDescriptorFactory.HUE_AZURE));
                    pilotMarker.setTitle("pilot " + id);
                    pilotMarker.setAlpha(0.5f);
                    markerPilot = pilotMarker;

                            /*googleMap.addMarker(
                            new MarkerOptions()
                                    .alpha(0.5f)
                                    .icon(BitmapDescriptorFactory.defaultMarker(BitmapDescriptorFactory.HUE_AZURE))
                                    .position(latLng)
                                    .title("pilot " + id));*/
                    markerPilot.setTag(new Pair<>(aircraft, this));
                }
                markerPilot.setPosition(latLng); // duplicate?
            }
        };

        @Override
        public void onChanged(@Nullable LocationData ignore) {
            boolean zoom = false;
            LocationData loc = aircraft.getLocation();
            if (loc == null || map == null || polylineOptions == null)
                return;

            // filter out zero data
            if (loc.getLatitude() == 0.0 && loc.getLongitude() == 0.0)
                return;

            LatLng latLng = new LatLng(loc.getLatitude(), loc.getLongitude());
            if (marker == null) {
                String id = "ID missing";
                if (aircraft.getIdentification() != null)
                    id = aircraft.getIdentification().getUasIdAsString();

                Marker aircraftMarker = new Marker((zzt) map);
                aircraftMarker.setPosition(latLng);
                aircraftMarker.setIcon(BitmapDescriptorFactory.defaultMarker(BitmapDescriptorFactory.HUE_AZURE));
                aircraftMarker.setTitle("aircraft " + id);
                aircraftMarker.setAlpha(0.5f);
                marker = aircraftMarker;

                        /*googleMap.addMarker(
                        new MarkerOptions()
                                .alpha(0.5f)
                                .position(latLng)
                                .title("aircraft " + id));*/
                marker.setTag(aircraft);
                zoom = true;
            }

            polylineOptions.add(latLng);
            if (polyline != null) {
                //polyline.remove(); found no equivalent method
                polyline = null;
            }
            //polyline = googleMap.addPolyline(polylineOptions);
            polyline = new Polyline();
            polyline.addPoint(new GeoPoint(latLng.latitude, latLng.longitude));
            // make polyline clickable
            // polyline.setPoints(); deprecated method

            map.getOverlayManager().add(polyline);


            marker.setPosition(latLng);
            if (zoom) {
                //googleMap.moveCamera(CameraUpdateFactory.newLatLng(latLng));
                map.getController().animateTo(new GeoPoint(latLng.latitude, latLng.longitude));
            }
        }
    }
}